<p align="center"><img src="./favicon.ico" alt="Viewer2d"></p>

<p align="center">A simple 2D viewer to load model with objects that you can select, highlight, hide...</p>

![CI](https://github.com/bimdata/viewer-2d/workflows/CI/badge.svg)

# Install

Use npm:

```
npm i @bimdata/viewer2d
```

# Usage

## Example

```html
<body style="background-color: #DDD">
  <canvas
    id="canvas"
    style="width: 800px; height: 600px; background-color: #FFF"
  ></canvas>
  <script type="module">
    import Viewer2D from "./lib_folder/viewer2D.esm.js";

    const canvas = document.getElementById("canvas");

    const viewer2d = Viewer2D({ canvas });

    const model = {
      bbox: [0, 0, 150, 150],
      objects: [
        {
          type: "testType1",
          area: 1,
          path: new Path2D("M0 0 v 50 h 50 v -50 Z"),
        },
        {
          type: "testType2",
          area: 2,
          path: new Path2D(
            "M 75 75 m -75, 0 a 75,75 0 1,0 150,0 a 75,75 0 1,0 -150,0"
          ),
        },
      ],
    };

    viewer2d.loadModel(model);

    viewer2d.camera.fitView(model.bbox);

    viewer2d.on("pick", ({ object, rightClick }) => {
      if (rightClick) {
        viewer2d.camera.fitView(object.bbox);
      } else {
        viewer2d.scene.setObjectsSelected([object.id], !object.selected);
      }
    });

    viewer2d.on("hover", ({ object }) =>
      viewer2d.scene.setObjectsHighlighted([object.id], true)
    );

    viewer2d.on("hover-out", ({ object }) =>
      viewer2d.scene.setObjectsHighlighted([object.id], false)
    );
  </script>
</body>
```

## Result

![simple example](./assets/simpleExample.gif)

# API

## Factory

Viewer2D is created using a factory:

```javascript
import Viewer2D from "@bimdata/viewer2d"; // webpack / rollup
const canvas = document.getElementById("canvas");
const viewer2d = Viewer2D({ canvas });
```

`canvas` is mandatory and another properties can be passed to customized the viewer2d behavior:

| Property       | Type                | Description                                               |
| :------------- | :------------------ | :-------------------------------------------------------- |
| `canvas`       | `HTMLCanvasElement` | **Required**.                                             |
| `rotateKey`    | `string`            | **Default** = `"shift"`. The key used to rotate.          |
| `scaleSpeed`   | `number`            | **Default** = `0.002`. The wheel zoom speed.              |
| `rotateSpeed`  | `number`            | **Default** = `0.1`. The rotation speed.                  |
| `maxScale`     | `number`            | **Default** = `1000`. The max scale.                      |
| `minScale`     | `number`            | **Default** = `1`. The min scale.                         |
| `fitViewRatio` | `number`            | **Default** = `1`. The fit view ratio.                    |
| `arrowMove`    | `number`            | **Default** = `5`. The move triggered by keyboard arrows. |

The factory Change the objects visible property. the following properties:

| Property                                                           | Description                                                                                                                        |
| :----------------------------------------------------------------- | :--------------------------------------------------------------------------------------------------------------------------------- |
| `materials`                                                        | Returns all registered [`material`s](#Materials).                                                                                  |
| `loadModel(model: Model)`                                          | A function to load model into the viewer. Returns the model.                                                                       |
| `unloadModel(modelId: number)`                                     | A function to unload model. This removes all related objects. Returns true if the model is unloaded (if it was previously loaded). |
| `scene`                                                            | An object containing the displayed elements.                                                                                       |
| `scene.models`                                                     | Returns all loaded models.                                                                                                         |
| `scene.objects`                                                    | Returns all loaded objects.                                                                                                        |
| `scene.setObjectsVisible(ids: number[], visible: boolean)`         | Change the objects visible property.                                                                                               |
| `scene.setObjectsSelected(ids: number[], selected: boolean)`       | Change the objects selected property.                                                                                              |
| `scene.setObjectsHighlighted(ids: number[], highlighted: boolean)` | Change the objects highlighted property.                                                                                           |
| `scene.setObjectsPickable(ids: number[], pickable: boolean)`       | Change the objects pickable property.                                                                                              |
| `camera`                                                           | An object containing the camera utilities.                                                                                         |
| `camera.fitView(bbox: number[x1, y1, x2, y2])`                     | Fit view on the bbox.                                                                                                              |
| `destroy()`                                                        | Clean destoy.                                                                                                                      |
| `on(event: string, callback: function)`                            | Register [`event`](#Events) callback. Returns an unique subscriptionId.                                                            |
| `once(event: string, callback: function)`                          | Register [`event`](#Events) callback once. Returns an unique subscriptionId.                                                       |
| `off(subscriptionId: number)`                                      | Unregister [`event`](#Events) callback.                                                                                            |

## Events

The following events are emitted on the viewer2d object :

| Name            | Payload                                                                       | Description                                       |
| :-------------- | :---------------------------------------------------------------------------- | :------------------------------------------------ |
| `tick`          | `dt`: number                                                                  | The time in milliseconds from the last rendering. |
| `model-loaded`  | [`model`](#Model)                                                             | A model is loaded.                                |
| `pick`          | {[`object`](#Object), position : {x: number, y: number}, rightClick: boolean} | An object is picked.                              |
| `pick-nothing`  | {position, rightClick}                                                        | Nothing is picked.                                |
| `hover`         | {[`object`](#Object), position}                                               | Hover an object.                                  |
| `hover-surface` | {[`object`](#Object), position}                                               | Continuous hover on object.                       |
| `hover-out`     | {[`object`](#Object), position}                                               | Hover leaves object.                              |
| `hover-off`     | {position}                                                                    | Hover nothing.                                    |

## Model

A model is an object that must have an objects property. This property must be an array of [`object`](#Object).

All other properties added to this object are kept and can be used when model is returned by `loadModel` or `scene.models`. (see [bbox](#Example))

## Object

| Property | Type                                                                | Description                                                                                 |
| :------- | :------------------------------------------------------------------ | :------------------------------------------------------------------------------------------ |
| `path`   | [`Path2D`](https://developer.mozilla.org/en-US/docs/Web/API/Path2D) | **Required**.                                                                               |
| `area`   | `number`                                                            | **Optional**. Object with small areas are displayed on top of others to allow picking them. |
| `type`   | `string`·                                                           | **Optional**. Type is used to display custom [`material`](#Materials).                      |

## Materials

Some materials are present by default :

```javascript
{
  default: {
    strokeStyle: {
      lineWidth: 1,
      color: new Uint8ClampedArray([0, 0, 0]),
      opacity: 1,
    },
  },
  highlighted: {
    strokeStyle: {
      lineWidth: 2,
      color: new Uint8ClampedArray([0, 0, 255]),
      opacity: 1,
    },
    fillStyle: {
      color: new Uint8ClampedArray([0, 0, 200]),
      opacity: 0.1,
    },
  },
  selected: {
    strokeStyle: {
      lineWidth: 3,
      color: new Uint8ClampedArray([0, 255, 0]),
      opacity: 1,
    },
    fillStyle: {
      color: new Uint8ClampedArray([0, 200, 0]),
      opacity: 0.4,
    },
  },
};
```

Is is possible to customize default materials or add custom ones to be displayed on scpecific object type :

```javascript
viewer2d.materials.default.strokeStyle.lineWidth = 0.01;
viewer2d.materials.selected.strokeStyle.lineWidth = 0.02;
viewer2d.materials.highlighted.strokeStyle.lineWidth = 0.02;

// displayed on object with type "someType"
viewer2d.materials.someType = {
  fillStyle: {
    color: new Uint8ClampedArray([0, 255, 255]),
    opacity: 0.5,
  },
};
```

# Development

Build on change for development and serve:

```
npm run dev
```

E2e tests:

```
npm run build:prod
npm run test:e2e
```

To e2e test on development (build in production mode and open cypress instead of running it):

```
npm run test:e2e-dev
```

Lint:

```
npm run lint
```
