import { makeIdGenerator, capitalCaseFirstLetter } from "./utils.js";
import makeEventHandler from "./events.js";

export default () => {
  const events = makeEventHandler();

  const models = new Map();
  const objects = new Map();

  const modelIdGenerator = makeIdGenerator();
  const objectIdGenerator = makeIdGenerator();

  function loadModel(model) {
    if (
      !model ||
      typeof model !== "object" ||
      !model.objects ||
      typeof model.objects !== "object" ||
      !model.objects.length
    ) {
      throw 'model should be an object with an non empty array of "objects".';
    }
    model.id = modelIdGenerator();
    models.set(model.id, model);
    const newObjects = [];
    model.objects.forEach(object => {
      addMissingObjectBasicProperties(object);
      object.modelId = model.id;
      object.id = objectIdGenerator();
      objects.set(object.id, object);
      newObjects.push(object);
    });
    events.emit("objects-added", newObjects);
    events.emit("model-loaded", model);
  }

  function unloadModel(modelId) {
    const model = models.get(modelId);
    if (model) {
      model.objects.forEach(object => objects.delete(object.id));
      events.emit("objects-removed", model.objects);
      models.delete(modelId);
      events.emit("model-unloaded", model);
      return true;
    } else {
      return false;
    }
  }

  function getObject(objectId) {
    const object = Array.from(objects.values()).find(
      object => object.id === objectId
    );
    return object ? Object.freeze({ ...object }) : undefined;
  }

  function getObjects() {
    return Array.from(objects.values()).map(object =>
      Object.freeze({ ...object })
    );
  }

  function getModels() {
    return Array.from(models);
  }

  const makeObjectsPropertySetter = property => (ids = [], value) => {
    if (!Array.isArray(ids)) throw "Ids must be an Array";
    const updatedObjects = [];
    ids
      .map(id => objects.get(id))
      .filter(Boolean)
      .forEach(o => {
        if (o[property] !== value) {
          o[property] = value;
          updatedObjects.push(o);
        }
      });
    if (updatedObjects.length) {
      events.emit(`objects-${value ? "" : "un"}${property}`, updatedObjects);
    }
  };

  function makeObjectsPropertySetters(properties = []) {
    return properties.reduce((o, property) => {
      o[
        `setObjects${capitalCaseFirstLetter(property)}`
      ] = makeObjectsPropertySetter(property);
      return o;
    }, {});
  }

  return {
    loadModel,
    unloadModel,
    getModels,
    getObject,
    getObjects,
    ...makeObjectsPropertySetters([
      "visible",
      "selected",
      "highlighted",
      "pickable",
    ]),
    ...events,
  };
};

function addMissingObjectBasicProperties(object) {
  object.visible = object.visible === undefined ? true : object.visible;
  object.pickable = object.pickable === undefined ? true : object.pickable;

  object.selected = object.selected === undefined ? false : object.selected;
  object.highlighted =
    object.highlighted === undefined ? false : object.highlighted;
}
